"""Core commands for the application."""

##############################################################################
# Python imports.
from functools import partial

##############################################################################
# Textual imports.
from textual.command import DiscoveryHit, Hit, Hits, Provider

##############################################################################
# Backward-compatible typing.
from typing_extensions import Final


##############################################################################
class CoreCommands(Provider):
    """A source of commands for invoking the main application commands."""

    COMMANDS: Final[tuple[tuple[str, str, str], ...]] = (
        ("Help", "help", "Show the main help screen"),
        ("Visit Pinboard", "goto_pinboard", "Visit the Pinboard website"),
        (
            "Redownload/refresh bookmarks",
            "redownload",
            "Download a full fresh copy of your bookmarks from Pinboard",
        ),
        ("Search", "search", "Search for text in the currently-shown bookmarks"),
        (
            "Logout",
            "logout",
            "Forget your Pinboard API token and remove the local copies of all bookmarks.",
        ),
        (
            "Toggle details",
            "toggle_details",
            "Toggle the display of the details of the highlighted bookmark.",
        ),
    )

    async def discover(self) -> Hits:
        """Handle a request to discover commands.

        Yields:
            Command discovery hits for the command palette.
        """
        for command, action, help_text in self.COMMANDS:
            yield DiscoveryHit(
                command,
                partial(self.screen.run_action, action),
                help=help_text,
            )

    async def search(self, query: str) -> Hits:
        """Handle a request to search for commands that match the query.

        Args:
            query: The query from the user.

        Yields:
            Command hits for the command palette.
        """
        matcher = self.matcher(query)
        for command, action, help_text in self.COMMANDS:
            if match := matcher.match(command):
                yield Hit(
                    match,
                    matcher.highlight(command),
                    partial(self.screen.run_action, action),
                    help=help_text,
                )


### core_commands.py ends here



"""The main entry point for the application."""

##############################################################################
# Python imports.
from argparse import ArgumentParser, Namespace

##############################################################################
# Textual imports.
from textual import __version__ as __textual_version__

##############################################################################
# Local imports.
from . import __version__
from .app import Tinboard
from .data import ExitStates
from .widgets import Filters


##############################################################################
def get_args() -> Namespace:
    """Get the command line arguments.

    Returns:
        The parsed command line arguments.
    """
    parser = ArgumentParser(
        prog="tinboard",
        description="A terminal-based client for Pinboard.",
        epilog=f"v{__version__}",
    )

    # Add the add command
    parser.add_argument(
        "command",
        choices=["add"],
        nargs="?",
        help="A command to modify the startup behaviour",
    )

    # Add --filter
    parser.add_argument(
        "-f",
        "--filter",
        choices=Filters.core_filter_names(),
        default=Filters.core_filter_names()[0],
        help="Set the default filter to use on startup",
    )

    # Add --version
    parser.add_argument(
        "-v",
        "--version",
        help="Show version information.",
        action="version",
        version=f"%(prog)s {__version__} (Textual v{__textual_version__})",
    )

    # Return the arguments.
    return parser.parse_args()


##############################################################################
def run() -> None:
    """Run the application."""
    args = get_args()
    state = Tinboard(args.filter).run(inline=args.command == "add")
    if state == ExitStates.TOKEN_FORGOTTEN:
        if Tinboard.environmental_token():
            print(
                "It looks like your token is held in an environment variable. "
                "If you wish to have that forgotten you will need to remove it yourself."
            )
        else:
            print("The locally-held copy of your API token has been removed.")
    elif state == ExitStates.TOKEN_NEEDED:
        print("An API token is needed to be able to connect to Pinboard.")
    elif state == ExitStates.INLINE_SAVE_ERROR:
        print("There was an error writing the bookmark to the Pinboard server.")


##############################################################################
if __name__ == "__main__":
    run()

### __main__.py ends here



"""Dialog for checking if a bookmark is in the Wayback Machine."""

##############################################################################
# Python imports.
from datetime import datetime
from webbrowser import open as view_url

##############################################################################
# Humanize imports.
from humanize import naturaltime

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.reactive import var
from textual.screen import ModalScreen
from textual.widgets import Button, Label

##############################################################################
# Local imports.
from ..wayback import Availability, WaybackError, availability


##############################################################################
class WaybackChecker(ModalScreen[None]):
    """Dialog for checking if a URL is in the Wayback Machine."""

    CSS = """
    WaybackChecker {
        align: center middle;

        &> Vertical {
            height: auto;
            width: auto;
            background: $surface;
            border: panel $primary;
            border-title-color: $accent;

            Label {
                height: auto;
                padding: 1 1 0 1;
                margin-bottom: 1;
            }

            Horizontal {
                width: 100%;
                height: auto;
                align: center middle;
                #view {
                    margin-right: 1;
                }
            }

            LoadingIndicator {
                background: 0%;
            }
        }

        #view {
            display: none;
        }

        &.available {
            &> Vertical {
                border: panel $success;
                background: $success-darken-2;
                border-title-color: $text;
            }
            #view {
                display: block;
            }
        }

        &.not-available > Vertical {
            border: panel $warning;
            background: $warning-darken-2;
            border-title-color: $text;
        }

        &.error > Vertical {
            border: panel $error;
            background: $error-darken-2;
            border-title-color: $text;
        }

        &.loading > Vertical {
            min-width: 50vw;
            min-height: 15;
        }
    }
    """

    AUTO_FOCUS = "#close"

    BINDINGS = [("escape", "close")]

    _wayback_data: var[Availability | WaybackError | None] = var(None, init=False)
    """The wayback availability data.

    This can either be `Availability` data, a `WaybackError` if there was a
    problem, or `None`.
    """

    def __init__(self, url: str) -> None:
        """Initialise the dialog.

        Args:
            url: The URL to look up on the Wayback Machine.
        """
        super().__init__(classes="loading")
        self._url = url
        """The URL to check the Wayback Machine status for."""

    def on_mount(self) -> None:
        """Once the DOM is ready, start looking up the URL."""
        self.check_availability()

    def compose(self) -> ComposeResult:
        with Vertical(id="dialog") as dialog:
            dialog.loading = True
            dialog.border_title = "Wayback Machine"
            yield Label("Loading...")
            with Horizontal():
                yield Button("View archive", id="view")
                yield Button("Close", id="close")

    @work
    async def check_availability(self) -> None:
        """Check if the given URL is available in the Wayback Machine."""
        try:
            self._wayback_data = await availability(self._url)
        except WaybackError as error:
            self._wayback_data = error
        finally:
            self.set_class(False, "loading")
            self.query_one("#dialog").loading = False

    def _watch__wayback_data(self) -> None:
        """React to the data coming back from the Wayback Machine."""
        if self._wayback_data is None:
            return
        if isinstance(self._wayback_data, WaybackError):
            self.query_one(Label).update(
                f"Error talking to the Wayback Machine.\n\n{self._wayback_data}"
            )
            self.set_class(True, "error")
            self.query_one("#dialog", Vertical).set_class(True, "error")
            self.query_one("#close", Button).variant = "error"
            return
        if self._wayback_data.available:
            archived_on = datetime.strptime(
                self._wayback_data.timestamp, "%Y%m%d%H%M%S"
            )
            self.query_one(Label).update(
                "The bookmark is available in the Wayback Machine.\n\n"
                f"It was last archived {archived_on} with status "
                f"{self._wayback_data.status}.\n\n"
                f"That was {naturaltime(archived_on)}."
            )
            self.set_class(True, "available")
            self.query_one("#view", Button).variant = "success"
            self.query_one("#close", Button).variant = "success"
        else:
            self.query_one(Label).update(
                "The bookmark is not available in the Wayback Machine."
            )
            self.query_one("#close", Button).variant = "warning"
            self.set_class(True, "not-available")

    @on(Button.Pressed, "#view")
    def view_wayback_archive(self) -> None:
        """View the archive in the Wayback Machine."""
        if (
            isinstance(self._wayback_data, Availability)
            and self._wayback_data.available
        ):
            view_url(self._wayback_data.archive_url)

    @on(Button.Pressed, "#close")
    def action_close(self) -> None:
        """Close the help screen."""
        self.dismiss(None)


### wayback_checker.py ends here



"""The help screen of the application."""

##############################################################################
# Python imports.
from inspect import cleandoc
from typing import Any
from webbrowser import open as open_url

##############################################################################
# Textual imports.
from textual import __version__ as textual_version  # pylint:disable=no-name-in-module
from textual import on
from textual.app import ComposeResult
from textual.containers import Center, Vertical, VerticalScroll
from textual.screen import ModalScreen, Screen
from textual.widgets import Button, Markdown

##############################################################################
# Local imports.
from .. import __version__

##############################################################################
# The help text.
HELP = f"""\
# Tinboard v{__version__}

## Introduction

`Tinboard` is a terminal-based client for [pinboard.in](https://pinboard.in/).

{{context_help}}

## Other

Tinboard was created by and is maintained by [Dave Pearson](https://www.davep.org/).


Tinboard is Free Software and can be [found on GitHub](https://github.com/davep/tinboard).


This version of Tinboard is using [Textual](https://textual.textualize.io/) v{textual_version}.

## Licence

Tinboard - A client for the Pinboard bookmarking service.[EOL]
Copyright (C) 2023-2024 Dave Pearson

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
"""


##############################################################################
class Help(ModalScreen[None]):
    """The help screen."""

    CSS = """
    Help {
        align: center middle;

        Vertical {
            width: 75%;
            height: 90%;
            background: $surface;
            border: panel $primary;
            border-title-color: $accent;
        }

        VerticalScroll {
            scrollbar-gutter: stable;
        }

        Center {
            height: auto;
            width: 100%;
            border-top: solid $primary;
            padding-top: 1;
        }
    }
    """

    BINDINGS = [("escape", "close")]

    def __init__(self, help_for: Screen[Any]) -> None:
        """Initialise the help screen.

        Args:
            help_for: The screen to show the help for.
        """
        super().__init__()
        self._context_help = "\n\n".join(
            cleandoc(getattr(helper, "CONTEXT_HELP"))
            for helper in reversed(
                (
                    help_for.focused if help_for.focused is not None else help_for
                ).ancestors_with_self
            )
            if hasattr(helper, "CONTEXT_HELP")
        ).strip()

    def compose(self) -> ComposeResult:
        """Compose the layout of the help screen."""
        with Vertical() as help_screen:
            help_screen.border_title = "Help"
            with VerticalScroll():
                yield Markdown(
                    HELP.replace("[EOL]", "  ").format(context_help=self._context_help)
                )
            with Center():
                yield Button("Okay [dim]\\[Esc]")

    @on(Button.Pressed)
    def action_close(self) -> None:
        """Close the help screen."""
        self.dismiss(None)

    @on(Markdown.LinkClicked)
    def visit(self, event: Markdown.LinkClicked) -> None:
        """Visit any link clicked in the help."""
        open_url(event.href)


### help.py ends here



"""Code relating to the application's configuration file."""

##############################################################################
# Python imports.
from dataclasses import asdict, dataclass
from functools import lru_cache
from json import dumps, loads
from pathlib import Path

##############################################################################
# Local imports.
from .locations import config_dir


##############################################################################
@dataclass
class Configuration:
    """The configuration data for the application."""

    dark_mode: bool = True
    """Should we run in dark mode?"""

    details_visible: bool = True
    """Show the details pane be visible?"""

    sort_tags_by_count: bool = False
    """Should the tag menu sort on count?"""


##############################################################################
def configuration_file() -> Path:
    """The path to the file that holds the application configuration.

    Returns:
        The path to the configuration file.
    """
    return config_dir() / "configuration.json"


##############################################################################
def save_configuration(configuration: Configuration) -> Configuration:
    """Save the given configuration.

    Args:
        The configuration to store.

    Returns:
        The configuration.
    """
    load_configuration.cache_clear()
    configuration_file().write_text(
        dumps(asdict(configuration), indent=4), encoding="utf-8"
    )
    return load_configuration()


##############################################################################
@lru_cache(maxsize=None)
def load_configuration() -> Configuration:
    """Load the configuration.

    Returns:
        The configuration.

    Note:
        As a side-effect, if the configuration doesn't exist a default one
        will be saved to storage.

        This function is designed so that it's safe and low-cost to
        repeatedly call it. The configuration is cached and will only be
        loaded from storage when necessary.
    """
    source = configuration_file()
    return (
        Configuration(**loads(source.read_text(encoding="utf-8")))
        if source.exists()
        else save_configuration(Configuration())
    )


### config.py ends here



"""Core filtering commands for the command palette."""

##############################################################################
# Python imports.
from functools import partial

##############################################################################
# Textual imports.
from textual.command import DiscoveryHit, Hit, Hits, Provider

##############################################################################
# Local imports.
from ..widgets import Filters


##############################################################################
class CoreFilteringCommands(Provider):
    """A source of commands for invoking the core filters."""

    async def discover(self) -> Hits:
        """Handle a request to discover commands.

        Yields:
            Command discovery hits for the command palette.
        """
        for command in Filters.OPTIONS:
            yield DiscoveryHit(
                f"Show {command}",
                partial(
                    self.screen.post_message,
                    Filters.core_filter_message(command),
                ),
                help=(
                    "Show all bookmarks"
                    if command == "All"
                    else f"Show all bookmarks that are {command.lower()}"
                ),
            )

    async def search(self, query: str) -> Hits:
        """Handle a request to search for commands that match the query.

        Args:
            query: The query from the user.

        Yields:
            Command hits for the command palette.
        """
        matcher = self.matcher(query)
        for command in Filters.OPTIONS:
            full_command = f"Show {command}"
            if match := matcher.match(full_command):
                yield Hit(
                    match,
                    matcher.highlight(full_command),
                    partial(
                        self.screen.post_message,
                        Filters.core_filter_message(command),
                    ),
                    help=(
                        "Show all bookmarks"
                        if command == "All"
                        else f"Show all bookmarks that are {command.lower()}"
                    ),
                )


### core_filters.py ends here



"""Defines exit states for the application."""

##############################################################################
# Python imports.
from enum import Enum


##############################################################################
class ExitStates(Enum):
    """Exit states for the application."""

    OKAY = 0
    """The general exit state."""

    TOKEN_NEEDED = 1
    """The application exited because a token is needed and none was given."""

    TOKEN_FORGOTTEN = 2
    """The application exited because the user forgot the token."""

    INLINE_SAVE_ERROR = 3
    """There was an error with an inline save."""


### exit_states.py ends here



"""A dialog that lets the user provide a token."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from webbrowser import open as open_url

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Input, Label


##############################################################################
class TokenInput(ModalScreen[str | None]):
    """A modal screen that lets the user enter an API token."""

    CSS = """
    TokenInput {
        align: center middle;

        &> Vertical {
            padding: 1 2;
            width: 40%;
            min-width: 70;
            height: auto;
            background: $surface;
            border: panel $error;
            border-title-color: $text;

            &> Horizontal {
                margin-top: 1;
                height: auto;
                align-horizontal: right;

                &> Horizontal {
                    margin-left: 1;
                    height: auto;

                }

                Button {
                    margin-right: 1;
                }
            }

            Label {
                width: 1fr;
                height: auto;
                padding-left: 1;
                padding-right: 1;
                margin-bottom: 1;
            }

        }
    }
    """

    BINDINGS = [("escape", "cancel"), ("f1", "get_token")]

    def compose(self) -> ComposeResult:
        """Compose the content of the screen."""
        with Vertical() as dialog:
            dialog.border_title = "API Token Required"
            yield Label(
                "To use the [i]pinboard.in[/] API you will need an API token. "
                "You can [@click=screen.get_token]get your API token from your "
                "account page on [i]pinboard.in[/][/]."
            )
            yield Input(placeholder="Paste the token here")
            with Horizontal():
                with Horizontal():
                    yield Button("Go to token [dim]\\[F1][/]", id="get-token")
                yield Button("Connect", id="connect", variant="primary")
                yield Button("Cancel [dim]\\[Esc][/]", id="cancel", variant="error")

    @on(Button.Pressed, "#connect")
    @on(Input.Submitted)
    def confirm(self) -> None:
        """React to the user confirming their input."""
        if token := self.query_one(Input).value.strip():
            self.dismiss(token)
        else:
            self.notify("Please provide a token", severity="error")

    @on(Button.Pressed, "#cancel")
    def action_cancel(self) -> None:
        """React to the user cancelling the dialog."""
        self.dismiss(None)

    @on(Button.Pressed, "#get-token")
    def action_get_token(self) -> None:
        """Open the page for getting an API token."""
        open_url("https://pinboard.in/settings/password")
        self.query_one(Input).focus()


### token_input.py ends here



"""Provides for the command palette."""

##############################################################################
# Local imports.
from .bookmarks import BookmarkCommands
from .core_commands import CoreCommands
from .core_filters import CoreFilteringCommands
from .tags import TagCommands

##############################################################################
# Public symbols.
__all__ = [
    "BookmarkCommands",
    "CoreCommands",
    "CoreFilteringCommands",
    "TagCommands",
]

### __init__.py ends here



"""A version of Textual's OptionList with some more navigation options."""

##############################################################################
# Textual imports.
from textual.binding import Binding
from textual.widgets import OptionList


##############################################################################
class OptionListEx(OptionList):
    """The Textual `OptionList` with more navigation keys."""

    BINDINGS = [
        Binding("s, j", "cursor_down", show=False),
        Binding("w, k", "cursor_up", show=False),
    ]


### extended_option_list.py ends here



"""A dialog for getting text to search for."""

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Input


##############################################################################
class SearchInput(ModalScreen[str]):
    """A modal screen to get search text from the user."""

    CSS = """
    SearchInput {
        align: center middle;

        Input, Input:focus {
            border: round $accent;
            width: 60%;
            padding: 1;
            height: auto;
        }
    }
    """

    BINDINGS = [("escape", "escape")]

    def compose(self) -> ComposeResult:
        """Compose the input dialog."""
        yield Input(placeholder="Case-insensitive text to look for in the bookmarks")

    @on(Input.Submitted)
    def search(self) -> None:
        """Perform the search."""
        self.dismiss(self.query_one(Input).value.strip())

    def action_escape(self) -> None:
        """Escape out without searching."""
        self.app.pop_screen()


### search_input.py ends here



"""The main screen for the application."""

##############################################################################
# Python imports.
from functools import partial
from webbrowser import open as open_url

##############################################################################
# Pyperclip imports.
from pyperclip import PyperclipException
from pyperclip import copy as to_clipboard

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.screen import Screen
from textual.widgets import Footer, Header, Rule

##############################################################################
# Local imports.
from .. import __version__
from ..commands import (
    BookmarkCommands,
    CoreCommands,
    CoreFilteringCommands,
    TagCommands,
)
from ..data import (
    ExitStates,
    bookmarks_file,
    load_configuration,
    save_configuration,
    token_file,
)
from ..messages import (
    AddBookmark,
    CheckWaybackMachine,
    ClearTags,
    CopyBookmarkURL,
    DeleteBookmark,
    EditBookmark,
    ShowAlsoTaggedWith,
    ShowTaggedWith,
    ToggleBookmarkPublic,
    ToggleBookmarkRead,
)
from ..pinboard import API, BookmarkData
from ..widgets import Bookmark, Bookmarks, Details, Filters, TagsMenu
from .bookmark_input import BookmarkInput
from .confirm import Confirm
from .help import Help
from .search_input import SearchInput
from .wayback_checker import WaybackChecker


##############################################################################
def filter_binding(name: str) -> Binding:
    """Create a binding for one of the core filters.

    Args:
        name: The name of the filter.

    Returns:
        A binding for the filter.
    """
    return Binding(Filters.shortcut(name), f"show_{name.lower()}")


##############################################################################
class Main(Screen[None]):
    """The main application screen."""

    CONTEXT_HELP = """
    ## Application keys and commands

    The following keys and commands are available throughout the application:

    | Key | Command | Description |
    | - | - | - |
    | <kbd>F1</kbd> | `Help` | This help screen. |
    | <kbd>F2</kbd> | `Visit Pinboard` | Visit the main Pinboard website. |
    | <kbd>F3</kbd> | | Toggle the bookmark details pane. |
    | <kbd>F4</kbd> | | Toggle the sort order of the tags menu. |
    | <kbd>F12</kbd> | `Logout` | Forgot your API token and remove the local bookmark cache. |
    | <kbd>Ctrl</kbd>+<kbd>l</kbd> | `Redownload/refresh bookmarks` | Reload the local bookmarks from Pinboard. |
    | <kbd>Ctrl</kbd>+<kbd>q</kbd> | `Quit the application` | Shockingly... quit the application! |
    | <kbd>Ctrl</kbd>+<kbd>p</kbd> | | Show the command palette. |
    | <kbd>/</kbd> | | Text search. |
    | <kbd>#</kbd> | | Focus the menu of tags. |
    | <kbd>a</kbd> | `Show All` | Show all bookmarks. |
    | <kbd>R</kbd> | `Show Unread` | Show all unread bookmarks. |
    | <kbd>r</kbd> | `Show Read` | Show all read bookmarks. |
    | <kbd>P</kbd> | `Show Private` | Show all private bookmarks. |
    | <kbd>p</kbd> | `Show Public` | Show all public bookmarks. |
    | <kbd>T</kbd> | `Show Untagged` | Show all untagged bookmarks. |
    | <kbd>t</kbd> | `Show Tagged` | Show all tagged bookmarks. |
    """

    TITLE = f"Tinboard v{__version__}"
    SUB_TITLE = "A pinboard.in client"
    COMMANDS = {
        BookmarkCommands,
        CoreCommands,
        CoreFilteringCommands,
        TagCommands,
    }

    CSS = """
    Main {
        layout: horizontal;
    }

    Header {
        HeaderIcon {
            visibility: hidden;
        }

        &.-tall {
            height: 1;
        }
    }

    Main > .focus {
        border: none;
        border-left: tall $accent 50%;
        background: $boost;
    }

    Main > .focus:focus, Main > .focus:focus-within {
        border: none;
        border-left: tall $accent;
        background: $panel;
    }

    #menu {
        padding: 0;
        margin: 0;
        width: 2fr;
        height: 1fr;
        min-width: 28;

        Filters {
            padding-left: 1;
        }

        TagsMenu {
            height: 1fr;
        }

        Rule {
            height: 1;
            margin: 0 0 0 0;
            background: $boost;
            color: $accent 50%;
        }
    }

    #menu:focus-within Rule {
        background: $boost;
        color: $accent;
    }

    Bookmarks {
        height: 1fr;
        width: 5fr;
    }

    Details {
        height: 1fr;
        width: 3fr;
        min-width: 30;
    }

    /* Tweaks to the above when the details are hidden. */

    Main.details-hidden Details {
        display: none;
    }

    Main.details-hidden Bookmarks {
        width: 8fr;
    }
    """

    BINDINGS = [
        filter_binding("All"),
        filter_binding("Public"),
        filter_binding("Private"),
        filter_binding("Unread"),
        filter_binding("Read"),
        filter_binding("Tagged"),
        filter_binding("Untagged"),
        Binding("f1", "help", "Help"),
        Binding("f2", "goto_pinboard"),
        Binding("f3", "toggle_details"),
        Binding("f4", "toggle_tag_order"),
        Binding("f12", "logout"),
        Binding("ctrl+l", "redownload"),
        Binding("escape", "escape"),
        Binding("ctrl+q", "quit", "Quit"),
        Binding("#", "focus('tags-menu')"),
        Binding("/", "search"),
    ]

    def __init__(
        self,
        api: API,
        default_filters: set[type[Filters.CoreFilter]] | None = None,
    ) -> None:
        """Initialise the main screen.

        Args:
            api: The Pinboard API.
            default_filters: Any default filters to apply on startup.
        """
        super().__init__()
        self._api = api
        """The Pinboard API client."""
        self._default_filters: set[type[Filters.CoreFilter]] = default_filters or set()
        """The default filters to use on startup."""

    def compose(self) -> ComposeResult:
        """Lay out the content of the screen."""
        yield Header()
        with Vertical(id="menu", classes="focus"):
            yield Filters()
            yield Rule()
            yield TagsMenu(id="tags-menu")
        yield Bookmarks(classes="focus")
        yield Details(classes="focus")
        yield Footer()

    def on_mount(self) -> None:
        """Start the process of loading the bookmarks."""
        configuration = load_configuration()
        self.set_class(not configuration.details_visible, "details-hidden")
        self.sub_title = "Loading cached bookmarks..."
        self.query_one(Bookmarks).loading = True
        self.query_one(TagsMenu).sort_by_count = configuration.sort_tags_by_count
        self.load_bookmarks()
        for default_filter in self._default_filters:
            self.post_message(default_filter())

    @work(thread=True)
    def load_bookmarks(self) -> None:
        """Load the local copy of the bookmarks, if they exist."""
        self.query_one(Bookmarks).load()
        self.app.call_from_thread(self.maybe_redownload)

    @work
    async def download_bookmarks(self) -> None:
        """Get all the bookmarks from Pinboard.

        Note:
            As a side-effect of calling this method, the local copy of all
            the bookmarks will be overwritten.
        """
        try:
            await self.query_one(Bookmarks).download_all(self._api)
        except API.RequestError:
            self.app.bell()
            self.notify(
                "Error downloading bookmarks from the server.",
                title="Download Error",
                severity="error",
                timeout=8,
            )
            self._bookmarks_changed()
        except OSError as error:
            self.app.bell()
            self.notify(
                f"Error saving the bookmarks.\n\n{error}",
                title="Save Error",
                severity="error",
                timeout=8,
            )

    @work
    async def maybe_redownload(self) -> None:
        """Redownload the bookmarks if they look newer on the server."""
        if last_download := self.query_one(Bookmarks).bookmarks.last_downloaded:
            try:
                latest_on_server = await self._api.last_update()
            except API.Error:
                self.app.bell()
                self.notify(
                    "Unable to get the last change date from Pinboard. Is your token valid?",
                    title="Server Error",
                    severity="error",
                    timeout=8,
                )
                return
            if latest_on_server > last_download:
                self.notify(
                    "Bookmarks on the server appear newer; downloading a fresh copy."
                )
                self.action_redownload()
        else:
            self.notify("No local bookmarks found; checking with the server.")
            self.action_redownload()

    @on(Bookmarks.Changed)
    def _bookmarks_changed(self) -> None:
        """Refresh the display when an update happens."""
        bookmarks = self.query_one(Bookmarks)
        bookmarks.loading = False
        self.query_one(TagsMenu).show(bookmarks.tag_counts)
        TagCommands.current_tags = list(bookmarks.tags)
        self.query_one(Details).bookmark = bookmarks.current_bookmark
        # TODO: if getting the counts starts to look like it causes a wee
        # pause, perhaps calculate them from within a threaded worker.
        # Mostly though, so far, I'm not seeing any impact.
        self.query_one(Filters).counts = bookmarks.bookmarks.counts
        bookmarks.focus()

    @on(Bookmarks.OptionHighlighted, "Bookmarks")
    def refresh_details(self, event: Bookmarks.OptionHighlighted) -> None:
        """Show the details of a highlighted bookmark.

        Args:
            event: The event causing the refresh.
        """
        assert isinstance(event.option, Bookmark)
        self.query_one(Details).bookmark = event.option

    def action_help(self) -> None:
        """Show the help screen."""
        self.app.push_screen(Help(self))

    def action_goto_pinboard(self) -> None:
        """Open Pinbaord in the user's web browser."""
        open_url("https://pinboard.in")

    def action_toggle_details(self) -> None:
        """Toggle the display of the details pane."""
        self.toggle_class("details-hidden")
        config = load_configuration()
        config.details_visible = not self.has_class("details-hidden")
        save_configuration(config)

    def action_toggle_tag_order(self) -> None:
        """Toggle the ordering of the tags in the tag menu."""
        tags = self.query_one(TagsMenu)
        tags.sort_by_count = not tags.sort_by_count
        tags.show(self.query_one(Bookmarks).tag_counts)
        config = load_configuration()
        config.sort_tags_by_count = tags.sort_by_count
        save_configuration(config)

    def _logout(self, confirmed: bool) -> None:
        """Process the logout confirmation.

        Args:
            confirmed: Was the logout confirmed by the user?
        """
        if confirmed:
            token_file().unlink(True)
            bookmarks_file().unlink(True)
            self.app.exit(ExitStates.TOKEN_FORGOTTEN)

    def action_logout(self) -> None:
        """Perform the logout action."""
        self.app.push_screen(
            Confirm(
                "Logout",
                "Remove the local copy of your API token and delete the local copy of all bookmarks?",
            ),
            callback=self._logout,
        )

    def action_redownload(self) -> None:
        """Freshly download the bookmarks."""
        self.sub_title = "Loading..."
        self.query_one(TagsMenu).show([])
        self.query_one(Bookmarks).loading = True
        self.download_bookmarks()

    def action_escape(self) -> None:
        """Give some context to banging the escape key."""
        if self.screen.focused is None:
            return
        if isinstance(self.screen.focused, Details) or isinstance(
            self.screen.focused.parent, Details
        ):
            self.query_one(Bookmarks).focus()
        elif isinstance(self.screen.focused, (Bookmarks, TagsMenu)):
            self.query_one(Filters).focus()
        elif isinstance(self.screen.focused, Filters):
            self.app.exit()

    @on(Filters.ShowAll)
    def action_show_all(self) -> None:
        """Show all bookmarks."""
        self.query_one(Bookmarks).show_all()

    @on(Filters.ShowPublic)
    def action_show_public(self) -> None:
        """Show all public bookmarks."""
        self.query_one(Bookmarks).public_filter = (
            None if self.query_one(Bookmarks).public_filter is True else True
        )

    @on(Filters.ShowPrivate)
    def action_show_private(self) -> None:
        """Show all private bookmarks."""
        self.query_one(Bookmarks).public_filter = (
            None if self.query_one(Bookmarks).public_filter is False else False
        )

    @on(Filters.ShowUnread)
    def action_show_unread(self) -> None:
        """Show all unread bookmarks."""
        self.query_one(Bookmarks).read_filter = (
            None if self.query_one(Bookmarks).read_filter is False else False
        )

    @on(Filters.ShowRead)
    def action_show_read(self) -> None:
        """Show all read bookmarks."""
        self.query_one(Bookmarks).read_filter = (
            None if self.query_one(Bookmarks).read_filter is True else True
        )

    @on(Filters.ShowUntagged)
    def action_show_untagged(self) -> None:
        """Show all untagged bookmarks."""
        self.query_one(Bookmarks).has_tags_filter = (
            None if self.query_one(Bookmarks).has_tags_filter is False else False
        )

    @on(Filters.ShowTagged)
    def action_show_tagged(self) -> None:
        """Show all tagged bookmarks."""
        self.query_one(Bookmarks).has_tags_filter = (
            None if self.query_one(Bookmarks).has_tags_filter is True else True
        )

    def _search(self, search_text: str) -> None:
        """Handle a request to search for text in the bookmarks."""
        self.query_one(Bookmarks).text_filter = search_text

    def action_search(self) -> None:
        """Do some free-text searching."""
        self.app.push_screen(SearchInput(), callback=self._search)

    @on(ShowTaggedWith)
    def show_tagged_with(self, event: ShowTaggedWith) -> None:
        """Show all bookmarks tagged with a given tag.

        Args:
            event: The event that contains the tag to show.
        """
        self.query_one(Bookmarks).tag_filter = {event.tag}

    @on(ShowAlsoTaggedWith)
    def show_also_tagged_with(self, event: ShowAlsoTaggedWith) -> None:
        """Add a tag to any current tag filter and show the matching bookmarks.

        Args:
            event: The event that contains the tag to add.
        """
        self.query_one(Bookmarks).tag_filter |= {event.tag}

    @on(ClearTags)
    def clear_tags(self) -> None:
        """Clear any tags that are in use."""
        self.query_one(Bookmarks).tag_filter = set()

    async def post_result(self, result: BookmarkData | None) -> None:
        """Handle the result of an edit of a bookmark.

        Args:
            result: The result data, or `None` if the edit was cancelled.
        """
        if result:
            try:
                await self._api.add_bookmark(result)
                self.query_one(Bookmarks).update_bookmark(result).save()
                self.notify("Bookmark saved.")
            except API.Error as error:
                self.app.bell()
                self.notify(
                    str(error),
                    title="Error saving bookmark data",
                    severity="error",
                    timeout=10,
                )
                self.app.push_screen(
                    BookmarkInput(
                        self._api,
                        result,
                        known_tags=self.query_one(Bookmarks).bookmarks.tags,
                    ),
                    callback=self.post_result,
                )
            except OSError as error:
                self.app.bell()
                self.notify(
                    f"Error saving the bookmarks.\n\n{error}",
                    title="Save Error",
                    severity="error",
                    timeout=8,
                )

    @on(CopyBookmarkURL)
    def copy_bookmark_to_clipboard(self) -> None:
        """Copy the currently-highlighted bookmark to the clipboard."""
        if (bookmark := self.query_one(Bookmarks).current_bookmark) is None:
            self.app.bell()
        elif bookmark.data.href:
            # Use Textual's own "copy to clipboard" as this will help push
            # it through remote connections.
            self.app.copy_to_clipboard(bookmark.data.href)
            # Having done that, try and use pyperclip anyway, because it's
            # possible we're within a terminal that doesn't support the
            # above, but this will at best finally get it there, and at
            # worst overwrite the clipboard with the same data.
            try:
                to_clipboard(bookmark.data.href)
            except PyperclipException:
                self.app.bell()
                self.notify(
                    "Clipboard support not available in your environment.",
                    severity="error",
                )
            else:
                self.notify("URL copied to the clipboard")

    @on(CheckWaybackMachine)
    def check_wayback(self) -> None:
        """Check if the currently-highlighted bookmark is archives."""
        if (bookmark := self.query_one(Bookmarks).current_bookmark) is None:
            self.app.bell()
        elif bookmark.data.href:
            self.app.push_screen(WaybackChecker(bookmark.data.href))

    @on(AddBookmark)
    def add(self) -> None:
        """Add a new bookmark."""
        self.app.push_screen(
            BookmarkInput(
                self._api, known_tags=self.query_one(Bookmarks).bookmarks.tags
            ),
            callback=self.post_result,
        )

    @on(EditBookmark)
    def edit(self) -> None:
        """Edit the current bookmark, if there is one."""
        if (bookmark := self.query_one(Bookmarks).current_bookmark) is None:
            self.app.bell()
        else:
            self.app.push_screen(
                BookmarkInput(
                    self._api,
                    bookmark.data,
                    known_tags=self.query_one(Bookmarks).bookmarks.tags,
                ),
                callback=self.post_result,
            )

    async def _delete(self, bookmark: Bookmark, confirmed: bool) -> None:
        """Respond to the user's confirmation about a bookmark deletion.

        Args:
            bookmark: The bookmark to delete.
            confirmed: The decision the user made about deletion.
        """
        if confirmed:
            try:
                await self._api.delete_bookmark(bookmark.data.href)
            except API.Error:
                self.app.bell()
                self.notify(
                    "Error trying to delete the bookmark.",
                    title="Server Error",
                    severity="error",
                    timeout=8,
                )
                return
            try:
                self.query_one(Bookmarks).remove_bookmark(bookmark).save()
            except OSError as error:
                self.app.bell()
                self.notify(
                    f"Error saving the bookmarks.\n\n{error}",
                    title="Save Error",
                    severity="error",
                    timeout=8,
                )
                return
            self.notify("Bookmark deleted.", severity="warning")

    @on(DeleteBookmark)
    def delete(self) -> None:
        """Delete the current bookmark, if there is one."""
        if (bookmark := self.query_one(Bookmarks).current_bookmark) is None:
            self.app.bell()
        else:
            self.app.push_screen(
                Confirm(
                    "Delete?",
                    f"Are you sure you wish to delete this bookmark?\n\n[dim i]{bookmark.data.description}[/]",
                ),
                callback=partial(self._delete, bookmark),
            )

    @on(ToggleBookmarkRead)
    async def toggle_read(self) -> None:
        """Toggle the read/unread status of the current bookmark."""
        if (bookmark := self.query_one(Bookmarks).current_bookmark) is None:
            self.app.bell()
        else:
            bookmark.data.to_read = not bookmark.data.to_read
            await self.post_result(bookmark.data)

    @on(ToggleBookmarkPublic)
    async def toggle_public(self) -> None:
        """Toggle the public/private status of the current bookmark."""
        if (bookmark := self.query_one(Bookmarks).current_bookmark) is None:
            self.app.bell()
        else:
            bookmark.data.shared = not bookmark.data.shared
            await self.post_result(bookmark.data)


### main.py ends here



"""Commands related to the list of bookmarks."""

##############################################################################
# Python imports.
from functools import partial

##############################################################################
# Textual imports.
from textual.command import DiscoveryHit, Hit, Hits, Provider

##############################################################################
# Backward-compatible typing.
from typing_extensions import Final

##############################################################################
# Local imports.
from ..messages import (
    AddBookmark,
    Command,
    CopyBookmarkURL,
    DeleteBookmark,
    EditBookmark,
    ToggleBookmarkPublic,
    ToggleBookmarkRead,
)


##############################################################################
class BookmarkCommands(Provider):
    """A source of commands for doing things with bookmarks."""

    COMMANDS: Final[tuple[tuple[str, type[Command], str], ...]] = (
        (
            "Add a new bookmark",
            AddBookmark,
            "Add a new bookmark to your bookmark collection",
        ),
        (
            "Copy to clipboard",
            CopyBookmarkURL,
            "Copy the URL for the current bookmark to the clipboard",
        ),
        ("Edit bookmark", EditBookmark, "Edit the current bookmark"),
        (
            "Toggle public/private",
            ToggleBookmarkPublic,
            "Toggle the current bookmark's public/private status",
        ),
        ("Delete bookmark", DeleteBookmark, "Delete the current bookmark"),
        (
            "Toggle read/unread",
            ToggleBookmarkRead,
            "Toggle the current bookmark's read/unread status",
        ),
    )

    async def discover(self) -> Hits:
        """Handle a request to discover commands.

        Yields:
            Command discovery hits for the command palette.
        """
        for command, message, help_text in self.COMMANDS:
            yield DiscoveryHit(
                command,
                partial(self.screen.post_message, message()),
                help=help_text,
            )

    async def search(self, query: str) -> Hits:
        """Handle a request to search for commands that match the query.

        Args:
            query: The query from the user.

        Yields:
            Command hits for the command palette.
        """
        matcher = self.matcher(query)
        for command, message, help_text in self.COMMANDS:
            if match := matcher.match(command):
                yield Hit(
                    match,
                    matcher.highlight(command),
                    partial(self.screen.post_message, message()),
                    help=help_text,
                )


### bookmarks.py ends here



"""The screens for the application."""

##############################################################################
# Local imports.
from .bookmark_input import BookmarkInput
from .help import Help
from .main import Main
from .token_input import TokenInput
from .wayback_checker import WaybackChecker

##############################################################################
# Exports.
__all__ = ["BookmarkInput", "Help", "Main", "TokenInput", "WaybackChecker"]

### __init__.py ends here



"""Functions for getting the locations of data."""

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# XDG imports.
from xdg_base_dirs import xdg_config_home, xdg_data_home


##############################################################################
def _tinboard_dir(root: Path) -> Path:
    """Given a root, ensure and return the Tinboard directory within it."""
    (save_to := root / "tinboard").mkdir(parents=True, exist_ok=True)
    return save_to


##############################################################################
def data_dir() -> Path:
    """The path to the data directory for the application.

    Returns:
        The path to the data directory for the application.

    Note:
        If the directory doesn't exist, it will be created as a side-effect
        of calling this function.
    """
    return _tinboard_dir(xdg_data_home())


##############################################################################
def config_dir() -> Path:
    """The path to the configuration directory for the application.

    Returns:
        The path to the configuration directory for the application.

    Note:
        If the directory doesn't exist, it will be created as a side-effect
        of calling this function.
    """
    return _tinboard_dir(xdg_config_home())


### locations.py ends here



"""Pinboard API code."""

##############################################################################
# Python imports.
from dataclasses import dataclass
from datetime import datetime, timedelta
from functools import reduce
from json import loads
from operator import or_
from typing import Final

##############################################################################
# HTTPX imports.
import httpx

##############################################################################
# pytz imports.
from pytz import UTC

##############################################################################
# Local imports.
from .bookmark_data import BookmarkData
from .parse_time import parse_time


##############################################################################
class API:
    """Pinboard API client class."""

    AGENT: Final[str] = "Tinboard (https://github.com/davep/tinboard)"
    """The agent string to use when talking to the API."""

    class Error(Exception):
        """Exception for any sort of error calling on the API."""

    class RequestError(Error):
        """An exception raised if there was a problem making the request."""

    class PinboardError(Error):
        """An exception raised if the Pinboard API raised an error."""

    def __init__(self, api_token: str) -> None:
        """Initialise the API object.

        Args:
            api_token: The API token to use.
        """
        self._api_token = api_token
        """The API token for the user."""
        self._all_cache: list[BookmarkData] = []
        """The cache of all bookmarks; used if we've grabbed them and we're calling for them again too soon."""
        self._last_all: datetime | None = None
        """The time when we last asked for all the bookmarks."""

    def _api_url(self, *path: str) -> str:
        """Construct a URL for calling on the API.

        Args:
            *path: The path to the endpoint.

        Returns:
            The URL to use.
        """
        return f"https://api.pinboard.in/v1/{'/'.join(path)}?auth_token={self._api_token}&format=json"

    async def _call(self, *path: str, **params: str) -> str:
        """Call on the Pinboard API.

        Args:
            path: The path for the API call.
            params: The parameters for the call.

        Returns:
            The text returned from the call.

        Raises:
            API.RequestError: If there was an error making the request.
        """
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(
                    self._api_url(*path),
                    params=params,
                    headers={"user-agent": self.AGENT},
                )
            except httpx.RequestError as error:
                raise self.RequestError(str(error))
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as error:
                raise self.RequestError(str(error))
            return response.text

    @classmethod
    def _check(cls, data: dict[str, str]) -> dict[str, str]:
        """Check if there's any sort of problem reported in the data.

        Args:
            data: The data to check.

        Returns:
            The data.

        Raises:
            API.PinboardError: If a problem was found.
        """
        if "code" in data and data["code"] != "done":
            raise cls.PinboardError(data["code"])
        return data

    async def last_update(self) -> datetime:
        """Get when the bookmarks were last updated on Pinboard.

        Returns:
            The time when the last update took place.
        """
        return parse_time(loads(await self._call("posts", "update"))["update_time"])

    _ALL_LIMIT: Final[timedelta] = timedelta(minutes=5)
    """The rate limit for getting all bookmarks."""

    async def _all_too_soon(self) -> bool:
        """Is it too soon to be getting all the bookmarks?"""
        # It's not too soon if we haven't grabbed everything yet.
        if self._last_all is None:
            return False
        # If we're within the rate limit, we also say that's too soon.
        if (datetime.now(UTC) - self._last_all) < self._ALL_LIMIT:
            return True
        # Finally, it is too soon if the server hasn't been updated.
        if (await self.last_update()) < self._last_all:
            return True
        return False

    async def all_bookmarks(self) -> list[BookmarkData]:
        """Get all the bookmarks available on the server.

        Returns:
            All the bookmarks.
        """
        # If we're hitting the API way too soon after a previous request...
        if await self._all_too_soon():
            # ...just go with the cached version.
            return self._all_cache
        self._all_cache = [
            BookmarkData.from_json(bookmark)
            for bookmark in loads(await self._call("posts", "all"))
        ]
        self._last_all = datetime.now(UTC)
        return self._all_cache

    @dataclass
    class TagSuggestions:
        """Type of data that holds tag suggestions for a URL."""

        url: str
        """The URL that the suggestions are for."""
        popular: list[str]
        """The popular tags."""
        recommended: list[str]
        """The recommended tags."""

    async def suggested_tags(self, url: str) -> TagSuggestions:
        """Get the suggested tags for the given URL.

        Args:
            url: The URL to get the suggestions for.

        Returns:
            The suggested tags for the URL.
        """
        suggestions = reduce(or_, loads(await self._call("posts", "suggest", url=url)))
        return self.TagSuggestions(
            url,
            popular=suggestions.get("popular", []),
            recommended=suggestions.get("recommended", []),
        )

    async def add_bookmark(self, data: BookmarkData) -> None:
        """Add the given bookmark to the server.

        Args:
            data: The bookmark to add.
        """
        self._check(
            loads(
                await self._call(
                    "posts",
                    "add",
                    url=data.href,
                    description=data.description,
                    extended=data.extended,
                    tags=data.tags,
                    dt=data.time.isoformat(),
                    replace="yes",
                    shared="yes" if data.shared else "no",
                    toread="yes" if data.to_read else "no",
                )
            )
        )

    async def delete_bookmark(self, url: str) -> None:
        """Delete the bookmark associated with a URL.

        Args:
            url: The URL to delete the bookmark for.
        """
        await self._call("posts", "delete", url=url)


### api.py ends here



"""Tag filtering commands for the command palette."""

##############################################################################
# Python imports.
from functools import partial

##############################################################################
# Textual imports.
from textual.command import DiscoveryHit, Hit, Hits, Provider

##############################################################################
# Local imports.
from ..messages import ClearTags, ShowAlsoTaggedWith, ShowTaggedWith


##############################################################################
class TagCommands(Provider):
    """A source of commands for filtering bookmarks by tag."""

    current_tags: list[str] = []
    """The current set of tags to draw on when setting up the commands."""

    async def discover(self) -> Hits:
        """Handle a request to discover commands.

        Yields:
            Command discovery hits for the command palette.
        """

        # Create commands for all the tags.
        for tag in self.current_tags:
            for prefix, message in (
                ("", ShowTaggedWith),
                ("Also", ShowAlsoTaggedWith),
            ):
                full_command = f"{prefix} tagged {tag}".strip().capitalize()
                yield DiscoveryHit(
                    full_command,
                    partial(self.screen.post_message, message(tag)),
                    help=f"Show all bookmarks that are {full_command.lower()}",
                )

        # Also add a command for clearing tags.
        yield DiscoveryHit(
            "Clear tag filter",
            partial(self.screen.post_message, ClearTags()),
            help="Clear any current tag-based filtering",
        )

    async def search(self, query: str) -> Hits:
        """Handle a request to search for commands that match the query.

        Args:
            query: The query from the user.

        Yields:
            Command hits for the command palette.
        """

        # Create commands for all the tags.
        matcher = self.matcher(query)
        for tag in self.current_tags:
            for prefix, message in (
                ("", ShowTaggedWith),
                ("Also", ShowAlsoTaggedWith),
            ):
                full_command = f"{prefix} tagged {tag}".strip().capitalize()
                if match := matcher.match(full_command):
                    yield Hit(
                        match,
                        matcher.highlight(full_command),
                        partial(self.screen.post_message, message(tag)),
                        help=f"Show all bookmarks that are {full_command.lower()}",
                    )

        # Also add a command for clearing tags.
        if match := matcher.match("Clear tag filter"):
            yield Hit(
                match,
                matcher.highlight("Clear tag filter"),
                partial(self.screen.post_message, ClearTags()),
                help="Clear any current tag-based filtering",
            )


### tags.py ends here



"""Provides a widget for displaying the bookmarks."""

##############################################################################
# Python imports.
from collections import Counter
from typing import Callable, Final, cast
from webbrowser import open as open_url

##############################################################################
# Humanize imports.
from humanize import naturaltime

##############################################################################
# Rich imports.
from rich.console import Group
from rich.emoji import Emoji
from rich.rule import Rule
from rich.table import Table

##############################################################################
# Textual imports.
from textual.app import DEFAULT_COLORS
from textual.binding import Binding
from textual.message import Message
from textual.reactive import var
from textual.widgets.option_list import Option, OptionDoesNotExist

##############################################################################
# Backward-compatible typing.
from typing_extensions import Self

##############################################################################
# Local imports.
from ..data import Bookmarks as LocalBookmarks
from ..messages import (
    AddBookmark,
    CheckWaybackMachine,
    CopyBookmarkURL,
    DeleteBookmark,
    EditBookmark,
    ToggleBookmarkPublic,
    ToggleBookmarkRead,
)
from ..pinboard import API, BookmarkData
from .extended_option_list import OptionListEx


##############################################################################
class Bookmark(Option):
    """An individual bookmark."""

    PRIVATE_ICON: Final[str] = Emoji.replace(":lock:")
    """The icon to use for a private bookmark."""

    UNREAD_ICON: Final[str] = Emoji.replace(":see-no-evil_monkey:")
    """The icon to use for an unread bookmark."""

    RULE: Final[Rule] = Rule(
        style=(
            DEFAULT_COLORS["dark"].accent.darken(0.15).rich_color.name
            if DEFAULT_COLORS["dark"].accent is not None
            else "blue"
        )
    )
    """The `Rule` to use between each bookmark."""

    def __init__(self, data: BookmarkData) -> None:
        """Initialise the bookmark.

        Args:
            data: The bookmark data gathered from the server.
        """
        self._data = data
        super().__init__(self.prompt, id=data.hash)

    @property
    def prompt(self) -> Group:
        """The prompt for the bookmark."""
        # Create the title and icons line.
        title = Table.grid(expand=True)
        title.add_column(ratio=1)
        title.add_column(justify="right")
        title.add_row(
            self._data.description,
            f"  {'' if self._data.shared else self.PRIVATE_ICON}{self.UNREAD_ICON if self._data.to_read else ''}",
        )
        # Create the details line.
        details = Table.grid(expand=True)
        details.add_column(ratio=1)
        details.add_column()
        details.add_row(
            f"[dim][i]{naturaltime(self._data.time)}[/][/]",
            f"[dim]{', '.join(sorted(self.data.tag_list, key = str.casefold))}[/]",
        )
        # Combine them and add a rule afterwards.
        return Group(title, details, self.RULE)

    @property
    def data(self) -> BookmarkData:
        """The bookmark as the underlying bookmark data."""
        return self._data

    @data.setter
    def data(self, data: BookmarkData) -> None:
        self._data = data
        self.set_prompt(self.prompt)


##############################################################################
class Bookmarks(OptionListEx):
    """The list of bookmarks."""

    CONTEXT_HELP = """
    ## Bookmarks keys and commands

    The following keys and commands are available in the bookmarks list:

    | Key | Command | Description |
    | - | - | - |
    | <kbd>Enter</kbd> | | Visit the current bookmark. |
    | <kbd>c</kbd> | `Copy to clipboard` | Copy the URL of the bookmark to the clipboard. |
    | <kbd>n</kbd> | `Add a new bookmark` | Create a new bookmark. |
    | <kbd>e</kbd> | `Edit bookmark` | Edit the currently-highlighted bookmark. |
    | <kbd>d</kbd> | `Delete bookmark` | Delete the currently-highlighted bookmark. |
    | <kbd>w</kbd> |  | Check if the bookmark is in the Wayback Machine. |
    | <kbd>Ctrl</kbd>+<kbd>r</kbd> | `Toggle read/unread` | Toggle the read/unread status of the currently-highlighted bookmark. |
    | <kbd>Ctrl</kbd>+<kbd>v</kbd> | `Toggle public/private` | Toggle the visibility of the currently-highlighted bookmark. |
    """

    DEFAULT_CSS = """
    Bookmarks {

        scrollbar-gutter: stable;

        .option-list--option {
            padding: 0 1 0 1;
        }
    }
    """

    BINDINGS = [
        Binding("enter", "visit", "Visit"),
        Binding("c", "copy"),
        Binding("n", "new", "New"),
        Binding("e", "edit", "Edit"),
        Binding("d", "delete", "Delete"),
        Binding("w", "check_wayback", "Wayback?"),
        Binding("ctrl+r", "read"),
        Binding("ctrl+v", "public"),
    ]

    bookmarks: var[LocalBookmarks] = var(LocalBookmarks, always_update=True, init=False)
    """The local copy of all the bookmarks."""

    read_filter: var[bool | None] = var(None, init=False)
    """The filter for the read status.

    `True` for read, `False` for unread, `None` for no filter.
    """

    public_filter: var[bool | None] = var(None, init=False)
    """The filter for the public status.

    `True` for public, `False` for private, `None` for no filter.
    """

    has_tags_filter: var[bool | None] = var(None, init=False)
    """The filter for if a bookmark has tags or not.

    `True` for has tags, `False` for doesn't have tags, `None` for no
    filter.
    """

    tag_filter: var[frozenset[str] | set[str]] = var(frozenset(), init=False)
    """The tags to filter on."""

    text_filter: var[str] = var("", init=False)
    """The text to filter on. Empty string for no filter."""

    _suspend_refresh: var[bool] = var(False)
    """Flag to say if a refresh should be suspended."""

    @property
    def highlighted_bookmark(self) -> Bookmark | None:
        """The currently-highlighted bookmark, if there is one."""
        if self.highlighted is not None:
            return cast(Bookmark, self.get_option_at_index(self.highlighted))
        return None

    def action_visit(self) -> None:
        """Visit the highlighted bookmark."""
        if (bookmark := self.highlighted_bookmark) is not None:
            if bookmark.data.href:
                open_url(bookmark.data.href)

    def action_copy(self) -> None:
        """Copy the URL of the current bookmark to the clipboard."""
        self.post_message(CopyBookmarkURL())

    def action_new(self) -> None:
        """Post the new bookmark command."""
        self.post_message(AddBookmark())

    def action_edit(self) -> None:
        """Post the edit command."""
        self.post_message(EditBookmark())

    def action_delete(self) -> None:
        """Post the delete command."""
        self.post_message(DeleteBookmark())

    def action_read(self) -> None:
        """Post the read status toggle command."""
        self.post_message(ToggleBookmarkRead())

    def action_public(self) -> None:
        """Post the public/private toggle command."""
        self.post_message(ToggleBookmarkPublic())

    async def action_check_wayback(self) -> None:
        """Check if the bookmark is in the Wayback Machine."""
        self.post_message(CheckWaybackMachine())

    @property
    def tags(self) -> list[str]:
        """All known tags in the current displayed set of bookmarks"""
        tags: set[str] = set()
        for n in range(self.option_count):
            bookmark = self.get_option_at_index(n)
            assert isinstance(bookmark, Bookmark)
            tags |= set(tag for tag in bookmark.data.tag_list)
        return sorted(list(tags), key=str.casefold)

    @staticmethod
    def _tag_key(tag: tuple[str, int]) -> str:
        """Get the key for a tag/count pair.

        Args:
            tag: The tag information.

        Returns:
            The key to use when sorting the key.
        """
        return tag[0].casefold()

    @property
    def tag_counts(self) -> list[tuple[str, int]]:
        """All known tags in the current displayed set of bookmarks, with counts."""
        tags: list[str] = []
        for n in range(self.option_count):
            bookmark = self.get_option_at_index(n)
            assert isinstance(bookmark, Bookmark)
            tags.extend(bookmark.data.tag_list)
        return sorted(list(Counter(tags).items()), key=self._tag_key)

    class Changed(Message):
        """Message to say the visible collection of bookmarks has changed."""

    def _refresh_bookmarks(self) -> Self:
        """Refresh the display of bookmarks.

        Takes core filters and tags into account.

        Returns:
            Self.
        """

        # GTFO if we're not supposed to refresh right now.
        if self._suspend_refresh:
            return self

        # Build up the filters.
        filter_names: list[str] = []
        filter_checks: list[Callable[[BookmarkData], bool]] = []
        if self.public_filter is not None:
            filter_names.append("Public" if self.public_filter else "Private")
            filter_checks.append(lambda bookmark: bookmark.shared is self.public_filter)
        if self.read_filter is not None:
            filter_names.append("Read" if self.read_filter else "Unread")
            filter_checks.append(
                lambda bookmark: bookmark.to_read is not self.read_filter
            )
        if self.has_tags_filter is not None:
            filter_names.append("Tagged" if self.has_tags_filter else "Untagged")
            filter_checks.append(
                lambda bookmark: bookmark.has_tags is self.has_tags_filter
            )
        if self.tag_filter:
            filter_names.append(f"Tagged {', '.join(sorted(self.tag_filter))}")
            filter_checks.append(lambda bookmark: bookmark.is_tagged(*self.tag_filter))
        if self.text_filter:
            filter_names.append(f"Contains '{self.text_filter}'")
            filter_checks.append(lambda bookmark: bookmark.has_text(self.text_filter))

        # Apply the filters.
        bookmarks = [
            bookmark for bookmark in self.bookmarks if bookmark.is_all(*filter_checks)
        ]

        # Set the title of the screen.
        # pylint:disable=attribute-defined-outside-init
        self.screen.sub_title = f"{'; '.join(filter_names) or 'All'} ({len(bookmarks)})"

        # Update the display of bookmarks, trying our best to keep the
        # currently-highlighted bookmark as the highlighted bookmark (if it
        # makes sense to, of course).
        highlighted_bookmark = (
            self.get_option_at_index(self.highlighted).id
            if self.highlighted is not None
            else None
        )
        try:
            return self.clear_options().add_options(
                Bookmark(bookmark) for bookmark in bookmarks
            )
        finally:
            self.post_message(self.Changed())
            if bookmarks:
                try:
                    self.highlighted = self.get_option_index(
                        highlighted_bookmark or "--none--"
                    )
                except OptionDoesNotExist:
                    self.highlighted = 0

    def show_all(self) -> None:
        """Show all bookmarks."""
        # There's currently no way to prevent watch methods from being
        # triggered while updating reactives; and in this one situation I
        # really could do with not having them fire as I want to set default
        # values for a lot of reactives but only refresh the bookmarks the
        # once. So here I implement my own flag to make that happen.
        self._suspend_refresh = True
        try:
            self.public_filter = None
            self.read_filter = None
            self.has_tags_filter = None
            self.text_filter = ""
            self.tag_filter = frozenset()
        finally:
            self._suspend_refresh = False
            self._refresh_bookmarks()

    def _watch_read_filter(self) -> None:
        """React to the read/unread filer being changed."""
        self._refresh_bookmarks()

    def _watch_public_filter(self) -> None:
        """React to the public bookmark filter being changed."""
        self._refresh_bookmarks()

    def _watch_has_tags_filter(self) -> None:
        """React to the has tags filter being changed."""
        self._refresh_bookmarks()

    def _watch_text_filter(self) -> None:
        """React to the search text being changed."""
        self._refresh_bookmarks()

    def _validate_tag_filter(
        self, new_value: frozenset[str] | set[str]
    ) -> frozenset[str]:
        """Ensure the tags filter always ends up being a frozen set."""
        return new_value if isinstance(new_value, frozenset) else frozenset(new_value)

    def _watch_tag_filter(self) -> None:
        """React to the tags being changed."""
        self._refresh_bookmarks()

    def _watch_bookmarks(self) -> None:
        """Refresh the display when all bookmarks are updated."""
        self._refresh_bookmarks()

    def load(self) -> Self:
        """Load the bookmarks from the local file.

        Returns:
            Self.
        """
        self.bookmarks = self.bookmarks.load()
        return self

    def save(self) -> Self:
        """Save the bookmarks to the local file."""
        self.bookmarks.save()
        return self

    async def download_all(self, api: API) -> Self:
        """Download all available bookmarks from the server.

        Args:
            api: The API to download via.

        Note:
            As a side-effect of calling this method, the local copy of all
            the bookmarks will be overwritten.

        Raises:
            API.RequestError: If there was an error downloading.
            OSError: If there was an error saving the bookmarks.
        """
        self.bookmarks = (await LocalBookmarks().download(api)).save()
        return self

    def update_bookmark(self, new_data: BookmarkData) -> Self:
        """Update the details of the given bookmark.

        Args:
            new_data: The new data for the bookmark.

        Returns:
            Self.
        """
        self.bookmarks.update_bookmark(new_data).mark_downloaded()
        return self._refresh_bookmarks()

    def remove_bookmark(self, bookmark: Bookmark) -> Self:
        """Remove the given bookmark.

        Args:
            bookmark: The bookmark to remove.

        Returns:
            Self.
        """
        self.bookmarks = self.bookmarks.remove_bookmark(bookmark.data).mark_downloaded()
        return self

    @property
    def current_bookmark(self) -> Bookmark | None:
        """The currently-highlighted bookmark, if there is one."""
        if (bookmark := self.highlighted) is None:
            return None
        return cast(Bookmark, self.get_option_at_index(bookmark))


### bookmarks.py ends here



"""Provides a function for parsing time."""

##############################################################################
# Python imports.
from datetime import datetime


##############################################################################
def parse_time(text: str) -> datetime:
    """Parse a time from the Pinboard API.

    Args:
        text: The text version of the time to parse.

    Returns:
        The parsed time.

    Pinboard returns times ending in a `Z`. Python doesn't seem capable of
    parsing that. So we swap that for a `+00:00` and then parse.
    """
    return datetime.fromisoformat(
        (text.removesuffix("Z") + "+00:00") if "Z" in text else text
    )


### parse_time.py ends here



"""Widgets used in the application."""

##############################################################################
# Local imports.
from .bookmarks import Bookmark, Bookmarks
from .details import Details
from .filters import Filters
from .tags import InlineTags, TagsMenu

##############################################################################
# Public symbols.
__all__ = [
    "Bookmark",
    "Bookmarks",
    "Details",
    "Filters",
    "InlineTags",
    "TagsMenu",
]

### __init__.py ends here



"""The main application class."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
import os

##############################################################################
# Textual imports.
from textual.app import App
from textual.binding import Binding

##############################################################################
# Local imports.
from .data import (
    Bookmarks,
    ExitStates,
    load_configuration,
    save_configuration,
    token_file,
)
from .pinboard import API, BookmarkData
from .screens import BookmarkInput, Main, TokenInput
from .widgets.filters import Filters


##############################################################################
class Tinboard(App[ExitStates]):
    """The application."""

    BINDINGS = [
        Binding("ctrl+backslash", "gndn"),
        Binding("ctrl+p", "command_palette", priority=True),
    ]

    def __init__(self, initial_filter: str | None = None) -> None:
        """Initialise the application.

        Args:
            initial_filter: The initial filter to apply when starting up.
        """
        super().__init__()
        self.dark = load_configuration().dark_mode
        self._initial_filter: set[type[Filters.CoreFilter]] = (
            set()
            if initial_filter is None
            else {Filters.core_filter_type(initial_filter)}
        )
        """The initial filter to apply when showing the main screen."""

    def token_bounce(self, token: str | None) -> None:
        """Handle the result of asking the user for their API token.

        Args:
            token: The resulting token.
        """
        if token:
            token_file().write_text(token, encoding="utf-8")
            self.push_screen(Main(API(token), self._initial_filter))
        else:
            self.exit(ExitStates.TOKEN_NEEDED)

    @staticmethod
    def environmental_token() -> str | None:
        """Try and get an API token from the environment.

        Returns:
           An API token found in the environment, or `None` if one wasn't found.
        """
        return os.environ.get("TINBOARD_API_TOKEN")

    @property
    def api_token(self) -> str | None:
        """The API token for talking to Pinboard.

        If the token is found in the environment, it will be used. If not a
        saved token will be looked for and used. If one doesn't exist the
        value will be `None`.
        """
        try:
            return self.environmental_token() or token_file().read_text(
                encoding="utf-8"
            )
        except IOError:
            pass
        return None

    async def inline_add(self, result: BookmarkData | None = None) -> None:
        """Handle the result adding a bookmark inline.

        Args:
            result: The result data, or `None` if the entry was cancelled.
        """
        if result:
            try:
                if token := self.api_token:
                    await API(token).add_bookmark(result)
            except API.Error:
                self.exit(ExitStates.INLINE_SAVE_ERROR)
        self.exit()

    def on_mount(self) -> None:
        """Display the main screen.

        Note:
            If the Pinboard API token isn't known, the token input dialog
            will first be shown; the main screen will then only be shown
            once the token has been acquired.
        """
        if token := self.api_token:
            if self.is_inline:
                # If we're running as an inline application, that means we
                # should simply be showing the bookmark input screen for
                # quick input. Note disabling the command palette too, which
                # has problems with inline mode.
                self.use_command_palette = False
                self.push_screen(
                    BookmarkInput(API(token), known_tags=Bookmarks().load().tags),
                    callback=self.inline_add,
                )
            else:
                # We're not in inline mode so we'll show the normal
                # application screen.
                self.push_screen(Main(API(token), self._initial_filter))
        else:
            self.push_screen(TokenInput(), callback=self.token_bounce)

    def _watch_dark(self) -> None:
        """Save the light/dark mode configuration choice."""
        configuration = load_configuration()
        configuration.dark_mode = self.dark
        save_configuration(configuration)


### app.py ends here



"""The details display widget."""

##############################################################################
# Python imports.
from webbrowser import open as open_url

##############################################################################
# Humanize imports.
from humanize import naturaltime

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import VerticalScroll
from textual.message import Message
from textual.reactive import var
from textual.widgets import Label

##############################################################################
# Local imports.
from ..messages import (
    CopyBookmarkURL,
    EditBookmark,
    ToggleBookmarkPublic,
    ToggleBookmarkRead,
)
from .bookmarks import Bookmark
from .tags import InlineTags


##############################################################################
class Link(Label):
    """Widget for showing the link.

    This is here mostly to work around the fact that a click action doesn't
    propagate in the way you'd expect.

    https://github.com/Textualize/textual/issues/3690
    """

    class Visit(Message):
        """Message to indicate that the link should be visited."""

    def action_visit(self) -> None:
        """Handle a UI request to visit the link."""
        self.post_message(self.Visit())


##############################################################################
class Details(VerticalScroll):
    """A widget for displaying details of a bookmark."""

    DEFAULT_CSS = """
    Details {
        scrollbar-gutter: stable;

        .hidden {
            visibility: hidden;
        }

        .empty {
            display: none;
        }

        Label {
            margin: 0 2 1 2;
            width: 1fr;
            color: $text;
        }

        #title {
            background: $primary;
            padding: 1 2 1 2;
            text-align: center;
        }

        .detail {
            background: $boost;
            padding: 1 2 1 2;
        }

        #added-ish {
            margin: 0 2 0 2;
            padding: 1 2 0 2;
        }

        #added-exact {
            margin: 0 2 1 2;
            padding: 0 2 1 2;
            text-align: right;
            color: $text-muted;
            text-style: italic;
        }

        InlineTags, InlineTags:focus {
            margin: 0 2 1 2;
        }
    }
    """

    BINDINGS = [
        Binding("enter", "visit_bookmark", "Visit"),
        Binding("c", "copy", "Copy to Clipboard"),
        Binding("e", "edit", "Edit"),
        Binding("ctrl+r", "read"),
        Binding("ctrl+v", "public"),
    ]

    CONTEXT_HELP = """
    ## Bookmark details keys

    The following keys are available in the bookmark details:

    | Key | Description |
    | - | - |
    | <kbd>Enter</kbd> | Visit the current bookmark. |
    | <kbd>c</kbd> | Copy the URL of the bookmark to the clipboard. |
    | <kbd>e</kbd> | Edit the details of the bookmark. |
    | <kbd>Ctrl</kbd>+<kbd>r</kbd> | Toggle the read/unread status of the bookmark. |
    | <kbd>Ctrl</kbd>+<kbd>v</kbd> | Toggle the visibility of the bookmark. |
    """

    bookmark: var[Bookmark | None] = var(None, always_update=True)
    """The current bookmark."""

    def compose(self) -> ComposeResult:
        """Compose the widget."""
        yield Label(id="title")
        yield Label(id="description", classes="detail empty")
        yield Link(id="link", classes="detail")
        yield Label(id="added-ish", classes="detail")
        yield Label(id="added-exact", classes="detail")
        yield Label(id="is-read", classes="detail")
        yield Label(id="is-public", classes="detail")
        yield InlineTags(classes="empty")

    def _watch_bookmark(self) -> None:
        """React to the bookmark being changed."""
        try:
            if self.bookmark is None:
                return
            self.query_one("#title", Label).update(self.bookmark.data.description)
            self.query_one("#description", Label).update(self.bookmark.data.extended)
            self.query_one("#description", Label).set_class(
                not bool(self.bookmark.data.extended), "empty"
            )
            self.query_one(Link).update(f"[@click=visit]{self.bookmark.data.href}[/]")
            self.query_one("#added-ish", Label).update(
                f"Added {naturaltime(self.bookmark.data.time)}"
            )
            self.query_one("#added-exact", Label).update(str(self.bookmark.data.time))
            self.query_one("#is-read", Label).update(
                f"The bookmark has {'[i]not[/] ' if self.bookmark.data.to_read else ''}been read"
            )
            self.query_one("#is-public", Label).update(
                f"The bookmark is {'[bold]public[/]' if self.bookmark.data.shared else '[dim]private[/]'}"
            )
            self.query_one(InlineTags).show(
                [
                    (tag, 1)
                    for tag in sorted(self.bookmark.data.tag_list, key=str.casefold)
                ]
            ).set_class(not self.bookmark.data.has_tags, "empty")

        finally:
            self.query("*").set_class(not bool(self.bookmark), "hidden")

    def action_edit(self) -> None:
        """Post the edit command."""
        self.post_message(EditBookmark())

    def action_read(self) -> None:
        """Post the read status toggle command."""
        self.post_message(ToggleBookmarkRead())

    def action_public(self) -> None:
        """Post the public/private toggle command."""
        self.post_message(ToggleBookmarkPublic())

    @on(Link.Visit)
    def action_visit_bookmark(self) -> None:
        """Visit the current bookmark, if there is one."""
        if self.bookmark is not None:
            if self.bookmark.data.href:
                open_url(self.bookmark.data.href)

    def action_copy(self) -> None:
        """Copy the URL of the bookmark to the clipboard."""
        self.post_message(CopyBookmarkURL())


### details.py ends here



"""Code relating to the data stored by the application."""

##############################################################################
# Local imports.
from .bookmarks import Bookmarks, bookmarks_file
from .config import load_configuration, save_configuration
from .exit_states import ExitStates
from .token import token_file

##############################################################################
# Exports.
__all__ = [
    "bookmarks_file",
    "Bookmarks",
    "ExitStates",
    "load_configuration",
    "save_configuration",
    "token_file",
]

### __init__.py ends here



"""Provides a suggester for tags."""

##############################################################################
# Python imports.
import re
from typing import Final, Iterable, Pattern

##############################################################################
# Textual imports.
from textual.suggester import Suggester


##############################################################################
class SuggestTags(Suggester):
    """A Textual `Input` suggester that suggests tags."""

    def __init__(self, tags: Iterable[str]) -> None:
        """Initialise the suggester.

        Args:
            tags: The collection of tags to suggest from.
        """
        # Note that we're going to be doing a case-insensitive match, but
        # the suggester API doesn't provided the raw value if you're not
        # being case-sensitive; so here we say we *are* going to be case
        # sensitive and then in get_suggestion we'll handle it ourselves.
        super().__init__(use_cache=True, case_sensitive=True)
        self._tags = list(tags)
        self._candidates = [candidate.casefold() for candidate in self._tags]

    _TAGS: Final[Pattern[str]] = re.compile(r"\s+")
    """Regular expression for splitting a string into individual tags."""

    _SUGGESTABLE: Final[Pattern[str]] = re.compile(r".*[^\s]$")
    """Regular expression to test if a value deserves a suggestion."""

    async def get_suggestion(self, value: str) -> str | None:
        """Get suggestions for the given value.

        Args:
            value: The value to make a suggestion for.

        Returns:
            A suggested completion, or `None` if none could be made.
        """
        if self._SUGGESTABLE.match(value):
            try:
                *other_tags, last_tag = [tag for tag in self._TAGS.split(value) if tag]
            except ValueError:
                return None
            used_tags = set(tag.casefold() for tag in other_tags)
            for candidate_index, candidate in enumerate(self._candidates):
                if candidate.startswith(last_tag) and candidate not in used_tags:
                    return value[: -len(last_tag)] + self._tags[candidate_index]
        return None


### tags.py ends here



"""Provides a class for holding data on a bookmark."""

##############################################################################
# Python imports.
from dataclasses import asdict, dataclass
from datetime import datetime
from hashlib import md5
from typing import Any, Callable

##############################################################################
# Local imports.
from .parse_time import parse_time


##############################################################################
@dataclass
class BookmarkData:
    """Pinboard bookmark data."""

    href: str
    """The URL for the bookmark."""

    description: str
    """The description (AKA title) of the bookmark."""

    extended: str
    """The extended description of the bookmark."""

    hash: str
    """The hash for the bookmark."""

    time: datetime
    """The time of the bookmark's modification."""

    shared: bool
    """Is the bookmark shared with the public?"""

    to_read: bool
    """Is the bookmark unread?"""

    tags: str
    """The tags for the bookmark, space-separated."""

    def __post_init__(self) -> None:
        """Final initialisation of the bookmark data.

        If `hash` is empty, one will be generated.
        """
        if not self.hash:
            self.hash = md5(self.href.encode()).hexdigest()

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> "BookmarkData":
        """Create an instance of `BookmarkData` from JSON data.

        Args:
            data: The data to create the bookmark instance from.

        Returns:
            An instance of `BookmarkData`.
        """

        def boolify(value: str | bool) -> bool:
            return value == "yes" if isinstance(value, str) else value

        # An accident of how this developed means that the Pinboard API
        # calls this `toread`, but I locally call it `to_read`, and I ended
        # up using the same code to pull the data back from JSON. So...
        if (to_read := "to_read") not in data:
            to_read = "toread"

        return cls(
            href=data.get("href", ""),
            description=data.get("description", ""),
            extended=data.get("extended", ""),
            hash=data.get("hash", ""),
            time=parse_time(data.get("time", "")),
            shared=boolify(data.get("shared", "")),
            to_read=boolify(data.get(to_read, "")),
            tags=data.get("tags", ""),
        )

    @property
    def as_json(self) -> dict[str, str]:
        """The bookmark in JSON-friendly form."""
        return asdict(self)

    @property
    def tag_list(self) -> list[str]:
        """The tags as a list of individual tags."""
        return self.tags.split()

    @property
    def has_tags(self) -> bool:
        """Does the bookmark have any tags?"""
        return bool(self.tag_list)

    def is_tagged(self, *tags: str) -> bool:
        """Is this bookmark tagged with the given tags?

        Args:
            tags: The tags to check for.

        Returns:
            `True` if the bookmark has all the tags, `False` if not.
        """
        return {tag.casefold() for tag in tags}.issubset(
            {tag.casefold() for tag in self.tag_list}
        )

    def has_text(self, search_text: str) -> bool:
        """Does the bookmark contain the given text?

        Note that this is a case-insensitive test.

        Args:
            search_text: The text to search for.

        Returns:
            `True` if the text was found anywhere in the bookmark, `False`
            if not.
        """
        return (
            search_text.casefold()
            in f"{self.description} {self.extended} {self.tags}".casefold()
        )

    def is_all(self, *checks: Callable[["BookmarkData"], bool]) -> bool:
        """Does this bookmark pass all the given tests?

        Args:
            checks: The checks to run against the bookmark.

        Returns:
            `True` if all tests pass, `False` if not.
        """
        return all(check(self) for check in checks)


### bookmark_data.py ends here



"""Generic custom messages for the application."""

##############################################################################
# Local imports.
from .commands import (
    AddBookmark,
    CheckWaybackMachine,
    Command,
    CopyBookmarkURL,
    DeleteBookmark,
    EditBookmark,
    ToggleBookmarkPublic,
    ToggleBookmarkRead,
)
from .tags import ClearTags, ShowAlsoTaggedWith, ShowTaggedWith

##############################################################################
# Exports.
__all__ = [
    "AddBookmark",
    "ClearTags",
    "Command",
    "CheckWaybackMachine",
    "CopyBookmarkURL",
    "EditBookmark",
    "DeleteBookmark",
    "ShowAlsoTaggedWith",
    "ShowTaggedWith",
    "ToggleBookmarkPublic",
    "ToggleBookmarkRead",
]

### __init__.py ends here



"""Defines the main filter menu for the application."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from typing import Final

##############################################################################
# Rich imports.
from rich.table import Table

##############################################################################
# Textual imports.
from textual import on
from textual.message import Message
from textual.reactive import var
from textual.widgets.option_list import Option

##############################################################################
# Local imports.
from ..data import Bookmarks
from .extended_option_list import OptionListEx


##############################################################################
class Filters(OptionListEx):
    """The main menu for the application."""

    DEFAULT_CSS = """
    Filters {
        height: auto;
        border: blank;

        .option-list--option {
            padding: 0 1;
        }

        &:focus {
            border: blank;
        }
    }
    """

    OPTIONS: Final[dict[str, str]] = {
        "All": "a",
        "Unread": "R",
        "Read": "r",
        "Private": "P",
        "Public": "p",
        "Untagged": "T",
        "Tagged": "t",
    }
    """The core options of the menu."""

    counts: var[Bookmarks.Counts] = var(Bookmarks.Counts(), init=False)
    """Holds the counts for the bookmarks."""

    @classmethod
    def shortcut(cls, option: str) -> str:
        """Get the shortcut for a given filter menu option.

        Args:
            option: The option to get the shortcut for.

        Returns:
            The shortcut for the option.
        """
        return cls.OPTIONS[option]

    def _prompt(self, name: str) -> Table:
        """Create a prompt for a filter.

        Args:
            name: The name of the filter.

        Returns:
            A renderable to use in the menu display.
        """
        prompt = Table.grid(expand=True)
        prompt.add_column(no_wrap=True, ratio=1)
        prompt.add_column(no_wrap=True, justify="left")
        prompt.add_column(no_wrap=True, justify="left")
        prompt.add_row(
            name,
            f"[dim i]{getattr(self.counts, name.lower())}[/]",
            f"  [dim]\\[{self.shortcut(name)}][/]",
        )
        return prompt

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,  # pylint:disable=redefined-builtin
        classes: str | None = None,
        disabled: bool = False,
    ):
        """Initialise the filter menu.

        Args:
            name: The name of the widget description.
            id: The ID of the widget description in the DOM.
            classes: The CSS classes of the widget description.
            disabled: Whether the widget description is disabled or not.
        """
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.add_options(
            [Option(self._prompt(option), id=option.lower()) for option in self.OPTIONS]
        )

    def on_focus(self) -> None:
        """Try and ensure an option is highlighted when we get focus."""
        if self.highlighted is None and self.option_count:
            self.highlighted = 0

    class CoreFilter(Message):
        """Base class for the core filters."""

    class ShowAll(CoreFilter):
        """Show all bookmarks."""

    class ShowUnread(CoreFilter):
        """Show unread bookmarks."""

    class ShowRead(CoreFilter):
        """Show read bookmarks"""

    class ShowPublic(CoreFilter):
        """Show the bookmarks that are public."""

    class ShowPrivate(CoreFilter):
        """Show the bookmarks that are private."""

    class ShowTagged(CoreFilter):
        """Show the bookmarks that have tags."""

    class ShowUntagged(CoreFilter):
        """Show the bookmarks that have no tags."""

    _CORE_FILTERS: Final[dict[str, type[CoreFilter]]] = {
        "all": ShowAll,
        "unread": ShowUnread,
        "read": ShowRead,
        "public": ShowPublic,
        "private": ShowPrivate,
        "tagged": ShowTagged,
        "untagged": ShowUntagged,
    }
    """The core filters and their message mappings."""

    @classmethod
    def core_filter_names(cls) -> list[str]:
        """The names for the core filters.

        Return:
            The names of the core filters.
        """
        return list(cls._CORE_FILTERS)

    @classmethod
    def core_filter_type(cls, name: str) -> type[CoreFilter]:
        """Get a core filter message type based off its name.

        Args:
            name: The name of the filter.

        Returns:
            A message type to request that filter be used.
        """
        return cls._CORE_FILTERS[name.lower()]

    @classmethod
    def core_filter_message(cls, name: str) -> CoreFilter:
        """Get a core filter message based off its name.

        Args:
            name: The name of the filter.

        Returns:
            A message to request that filter be used.
        """
        return cls.core_filter_type(name)()

    @on(OptionListEx.OptionSelected)
    def handle_selection(self, event: OptionListEx.OptionSelected) -> None:
        """Handle a menu option selection.

        Args:
            event: The event to handle.
        """
        event.stop()
        assert event.option.id is not None
        self.post_message(self.core_filter_message(event.option.id))

    def _watch_counts(self) -> None:
        """React to the counts being changed."""
        for filter_name in self.OPTIONS:
            self.replace_option_prompt(filter_name.lower(), self._prompt(filter_name))


### menu.py ends here



"""Provides a dialog for editing bookmark details."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from datetime import datetime
from urllib.parse import urlparse

##############################################################################
# Pyperclip imports.
from pyperclip import PyperclipException
from pyperclip import paste as from_clipboard

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.events import DescendantBlur, DescendantFocus
from textual.screen import ModalScreen
from textual.validation import Length
from textual.widgets import Button, Checkbox, Input, Label, TextArea

##############################################################################
# Local imports.
from ..pinboard import API, BookmarkData
from ..suggestions import SuggestTags


##############################################################################
class BookmarkInput(ModalScreen[BookmarkData | None]):
    """The bookmark editing dialog."""

    CSS = """
    BookmarkInput {
        align: center middle;

        Vertical {
            width: 60%;
            height: auto;
            background: $surface;
            border: panel $primary;
            border-title-color: $accent;

            &:inline {
                width: 100%;
                border: none;
            }
        }

        TextArea {
            padding: 0 2 0 2;
        }

        #description {
            height: 10;
        }

        #buttons {
            height: auto;
            margin-top: 1;
            align-horizontal: right;
        }

        Button {
            margin-right: 1;
        }

        Label {
            margin: 1 0 0 1;
        }

        #flags {
            height: auto;
            margin-top: 1;
        }

        #tag-suggestions {
            display: none;
            width: 1fr;

            &.got-suggestions {
                display: block;
            }
        }
    }
    """

    BINDINGS = [("escape", "cancel"), ("f2", "save")]

    def __init__(
        self,
        api: API,
        bookmark: BookmarkData | None = None,
        *,
        known_tags: list[str] | None = None,
    ) -> None:
        """Initialise the bookmark input dialog.

        Args:
            api: The Pinboard API object.
            bookmark: The bookmark to edit, or `None` for a new one.
            known_tags: The tags we currently know about.
        """
        super().__init__()
        self._api = api
        self._last_url = "" if bookmark is None else bookmark.href
        self._bookmark = bookmark
        self._tags = known_tags or []

    def compose(self) -> ComposeResult:
        """Compose the layout of the dialog."""
        with Vertical() as dialog:
            dialog.border_title = "Bookmark"
            yield Label("URL:")
            yield Input(
                placeholder="Bookmark URL", id="url", validators=[Length(minimum=1)]
            )
            yield Label("Title:")
            yield Input(placeholder="Bookmark title", id="title")
            yield Label("Description:")
            yield TextArea(id="description")
            yield Label("Tags:")
            yield Input(
                placeholder="Bookmark tags (space separated)",
                id="tags",
                suggester=SuggestTags(self._tags),
            )
            yield Label(id="tag-suggestions")
            with Horizontal(id="flags"):
                yield Checkbox("Private", id="private")
                yield Checkbox("Read Later", id="read-later")
            with Horizontal(id="buttons"):
                yield Button("Save [dim]\\[F2][/]", id="save")
                yield Button("Cancel [dim]\\[Esc][/]", id="cancel")

    @work(exclusive=True)
    async def _get_tag_suggestions(self) -> None:
        """Load up fresh tag suggestions based on the URL."""
        try:
            tags = await self._api.suggested_tags(self.query_one("#url", Input).value)
        except API.Error:
            self.app.bell()
            self.notify(
                "Error getting suggested tags from the server.",
                title="Server Error",
                severity="error",
                timeout=8,
            )
            return
        suggested_tags = ""
        if tags.recommended:
            suggested_tags += f"[b]Recommended:[/b] {' '.join(tags.recommended)}\n"
        if tags.popular:
            suggested_tags += f"[b]Popular:[/b] {' '.join(tags.popular)}"
        self.query_one("#tag-suggestions").set_class(
            bool(suggested_tags.strip()), "got-suggestions"
        )
        self.query_one("#tag-suggestions", Label).update(f"[dim]{suggested_tags}[/]")
        self.query_one("#tags", Input).suggester = SuggestTags(
            set(self._tags + tags.recommended + tags.popular)
        )

    @work(thread=True)
    def _suggest_bookmark(self) -> None:
        """Get a bookmark suggestion by peeking in the clipboard."""
        try:
            clipboard = from_clipboard()
        except PyperclipException:
            return
        if clipboard:
            if urlparse(clipboard).scheme in (  # pylint:disable=no-member
                "http",
                "https",
            ):

                def _paste() -> None:
                    url = self.query_one("#url", Input)
                    if not url.value:
                        url.value = clipboard
                        self._get_tag_suggestions()

                self.app.call_from_thread(_paste)

    def on_mount(self) -> None:
        """Configure the dialog once it's in the DOM."""
        if self._bookmark:
            self.query_one("#url", Input).value = self._bookmark.href
            self.query_one("#title", Input).value = self._bookmark.description
            self.query_one("#description", TextArea).text = self._bookmark.extended
            self.query_one("#tags", Input).value = self._bookmark.tags
            self.query_one("#private", Checkbox).value = not self._bookmark.shared
            self.query_one("#read-later", Checkbox).value = self._bookmark.to_read
            self._get_tag_suggestions()
        else:
            self._suggest_bookmark()

    @on(DescendantFocus, "#url")
    def _remember_url(self, event: DescendantFocus) -> None:
        """Save the URL on entry to the URL field.

        Args:
            event: The event to handle.
        """
        if isinstance(event.widget, Input):  # It should be, but help typing.
            self._last_url = event.widget.value.strip()

    @on(DescendantBlur, "#url")
    def _refresh_tag_suggestions(self, event: DescendantBlur) -> None:
        """Refresh the tag suggestions when leaving the URL field, having modified it."""
        if isinstance(event.widget, Input):  # It should be, but help typing.
            if (
                event.widget.value.strip()
                and event.widget.value.strip() != self._last_url
            ):
                self.query_one("#tag-suggestions").set_class(False, "got-suggestions")
                self._get_tag_suggestions()

    @on(Button.Pressed, "#save")
    def action_save(self) -> None:
        """Save the bookmark data."""
        if not self.query_one("#url", Input).value.strip():
            self.app.bell()
            self.notify(
                "The URL is required for the bookmark.",
                severity="error",
                title="Missing URL",
            )
            return
        self.dismiss(
            BookmarkData(
                href=self.query_one("#url", Input).value,
                description=self.query_one("#title", Input).value,
                extended=self.query_one("#description", TextArea).text,
                tags=self.query_one("#tags", Input).value,
                shared=not self.query_one("#private", Checkbox).value,
                to_read=self.query_one("#read-later", Checkbox).value,
                hash="" if self._bookmark is None else self._bookmark.hash,
                time=datetime.now() if self._bookmark is None else self._bookmark.time,
            )
        )

    @on(Button.Pressed, "#cancel")
    def action_cancel(self) -> None:
        """Cancel the edit of the bookmark data."""
        self.dismiss(None)


### bookmark_input.py ends here



"""Defines a widget for picking tags."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from typing import Final

##############################################################################
# Rich imports.
from rich.console import RenderableType
from rich.emoji import Emoji
from rich.table import Table

##############################################################################
# Textual imports.
from textual import on
from textual.binding import Binding
from textual.events import Focus
from textual.reactive import var
from textual.widgets.option_list import Option, OptionDoesNotExist

##############################################################################
# Backward-compatible typing.
from typing_extensions import Self

##############################################################################
# Local imports.
from ..messages import ClearTags, ShowAlsoTaggedWith, ShowTaggedWith
from .extended_option_list import OptionListEx


##############################################################################
class Tags(OptionListEx):
    """A menu of tags."""

    CONTEXT_HELP = """
    ## Tag list keys

    The following keys are available in the list of tags:

    | Key | Description |
    | - | - |
    | <kbd>Enter</kbd> | Show bookmarks with this tag in the bookmark list. |
    | <kbd>+</kbd> | Add this tag to any tag filter active in the bookmark list. |
    """

    DEFAULT_CSS = """
    Tags {
        &:focus {
            border: blank;
        }

        &> .option-list--option {
            padding: 0 1;
        }
    }
    """

    BINDINGS = [
        Binding("enter", "select", "Show tagged", show=True),
        Binding("+", "also_tagged", "Show also tagged"),
    ]

    def _prompt(self, tag: str, count: int) -> RenderableType:
        """A prompt for the given tag.

        Args:
            tag: The tag to build a prompt for.
            count: The count for that tag.

        Returns:
            The prompt for the tag.
        """
        prompt = Table.grid(expand=True)
        prompt.add_column(ratio=1)
        prompt.add_column(justify="right")
        prompt.add_row(tag, f"[dim i]{count}[/]")
        return prompt

    def _sorted(self, tags: list[tuple[str, int]]) -> list[tuple[str, int]]:
        """Sort the tags.

        Args:
            tags: The tags to sort.

        Returns:
            The tags in the desired sort order.
        """
        return tags

    def show(self, tags: list[tuple[str, int]]) -> Self:
        """Show the given list of tags.

        Args:
            tags: The tags to show in the widget.

        Returns:
            Self.
        """
        self.can_focus = bool(tags)
        highlighted_tag = (
            self.get_option_at_index(self.highlighted).id
            if self.highlighted is not None
            else None
        )
        try:
            return self.clear_options().add_options(
                [
                    Option(self._prompt(tag, count), id=tag)
                    for tag, count in self._sorted(tags)
                ]
            )
        finally:
            if tags:
                try:
                    self.highlighted = self.get_option_index(highlighted_tag or "")
                except OptionDoesNotExist:
                    self.highlighted = 0

    def _on_focus(self, _: Focus) -> None:
        """Highlight the first item on focus, if none highlighted."""
        if self.option_count and self.highlighted is None:
            self.highlighted = 0

    @on(OptionListEx.OptionSelected)
    def _show_tagged(self, event: OptionListEx.OptionSelected) -> None:
        """Request that bookmarks of a given tag are shown.

        Args:
            event: The event to handle.
        """
        if event.option.id is not None:
            self.post_message(ShowTaggedWith(event.option.id))

    def action_also_tagged(self) -> None:
        """Request that the current tag is added to any tag filter in place."""
        if self.highlighted is not None:
            if (tag := self.get_option_at_index(self.highlighted).id) is not None:
                self.post_message(ShowAlsoTaggedWith(tag))


##############################################################################
class TagsMenu(Tags):
    """A version of `Tags` to use as part of the main menu."""

    CONTEXT_HELP = f"{Tags.CONTEXT_HELP}| <kbd>c</kbd> | Clear any active tag filter. |"

    BINDINGS = [Binding("c", "clear", "Clear Tags")]

    sort_by_count: var[bool] = var(False)
    """Should the tags be sorted by count?"""

    def _sorted(self, tags: list[tuple[str, int]]) -> list[tuple[str, int]]:
        """Sort the tags.

        Args:
            tags: The tags to sort.

        Returns:
            The tags in the desired sort order.
        """
        return (
            sorted(tags, key=lambda tag: tag[1], reverse=True)
            if self.sort_by_count
            else tags
        )

    def action_clear(self) -> None:
        """Clear any active tags."""
        self.post_message(ClearTags())


##############################################################################
class InlineTags(Tags):
    """A version of the `Tags` widget intended to embed in another."""

    DEFAULT_CSS = """
    InlineTags > .option-list--option {
        padding: 0;
    }
    """

    _ICON: Final[str] = Emoji.replace(":bookmark: ")
    """The icon to show before tags."""

    def _prompt(self, tag: str, count: int) -> RenderableType:
        """A prompt for the given tag.

        Args:
            The tag to build a prompt for.

        Returns:
            The prompt for the tag.
        """
        del count
        return f"{self._ICON} {tag}"


### tags.py ends here



"""Code relating to the bookmarks file."""

##############################################################################
# Python imports.
from dataclasses import dataclass
from datetime import datetime
from json import JSONEncoder, dumps, loads
from pathlib import Path
from typing import Any, Iterator

##############################################################################
# pytz imports.
from pytz import UTC

##############################################################################
# Backward-compatible typing.
from typing_extensions import Self

##############################################################################
# Local imports.
from ..pinboard import API, BookmarkData
from .locations import data_dir


##############################################################################
def bookmarks_file() -> Path:
    """The path to the file that the local bookmarks are held in.

    Returns:
        The path to the bookmarks file.
    """
    return data_dir() / "bookmarks.json"


##############################################################################
class Bookmarks:
    """Class that holds the local copy of all the bookmarks."""

    def __init__(self) -> None:
        """Initialise the bookmarks."""
        super().__init__()
        self._bookmarks: list[BookmarkData] = []
        """The internal list of bookmarks."""
        self._index: dict[str, int] = {}
        """The index into the bookmarks."""
        self._last_downloaded: datetime | None = None
        """When the bookmarks were last downloaded."""

    @property
    def last_downloaded(self) -> datetime | None:
        """The time the bookmarks were last downloaded, or `None` if not yet."""
        return self._last_downloaded

    def _reindex(self) -> Self:
        """Reindex the bookmarks."""
        self._index = {
            bookmark.hash: index for index, bookmark in enumerate(self._bookmarks)
        }
        return self

    def mark_downloaded(self) -> Self:
        """Mark the bookmarks as having being downloaded at the time of calling."""
        self._last_downloaded = datetime.now(UTC)
        return self

    async def download(self, api: API) -> Self:
        """Download all available bookmarks from the server.

        Args:
            api: The API to download via.

        Returns:
            Self.
        """
        self._bookmarks = await api.all_bookmarks()
        return self._reindex().mark_downloaded()

    def _load_local_json(self, data: dict[str, Any]) -> Self:
        """Load the bookmarks from the given JSON data.

        Args:
            data: The data to load.

        Returns:
            Self.
        """
        self._last_downloaded = datetime.fromisoformat(data["last_downloaded"])
        self._bookmarks = [
            BookmarkData.from_json(bookmark) for bookmark in data["bookmarks"]
        ]
        return self._reindex()

    def load(self) -> Self:
        """Load the bookmarks from the local file.

        Returns:
            Self.
        """
        if bookmarks_file().exists():
            self._load_local_json(loads(bookmarks_file().read_text(encoding="utf-8")))
        return self

    class _Encoder(JSONEncoder):
        """Encoder for turning the bookmarks into JSON data."""

        def default(self, o: object) -> Any:
            """Handle unknown values.

            Args:
                o: The object to handle.
            """
            return datetime.isoformat(o) if isinstance(o, datetime) else o

    @property
    def _as_local_json(self) -> dict[str, Any]:
        """All the bookmarks as a JSON-friendly list."""
        return {
            "last_downloaded": None
            if self._last_downloaded is None
            else self._last_downloaded.isoformat(),
            "bookmarks": [bookmark.as_json for bookmark in self._bookmarks],
        }

    def save(self) -> Self:
        """Save the bookmarks to the local file.

        Returns:
            Self.
        """
        bookmarks_file().write_text(
            dumps(self._as_local_json, indent=4, cls=self._Encoder), encoding="utf-8"
        )
        return self

    def update_bookmark(self, bookmark: BookmarkData) -> Self:
        """Update the bookmarks with the given bookmark data.

        Args:
            bookmark: The bookmark data to update with.

        Returns:
            Self.

        Note:
            If the `hash` of the bookmark is known, the existing bookmark
            will be replaced with the new data. If the hash is unknown a
            brand new bookmark will be added to the collection.
        """
        # If the bookmark is already known...
        if (bookmark_index := self._index.get(bookmark.hash)) is not None:
            # ...replace it.
            self._bookmarks[bookmark_index] = bookmark
        else:
            # ...otherwise add it at the top of the list.
            self._bookmarks.insert(0, bookmark)
            self._reindex()
        return self

    def remove_bookmark(self, bookmark: BookmarkData) -> Self:
        """Remove the given bookmark data.

        Args:
            bookmark: The bookmark to remove.

        Returns:
            Self.
        """
        del self._bookmarks[self._index[bookmark.hash]]
        return self._reindex()

    @dataclass
    class Counts:
        """The various bookmark counts."""

        all: int = 0
        """The count of all the bookmarks."""
        unread: int = 0
        """The count of unread bookmarks."""
        read: int = 0
        """The count of read bookmarks."""
        private: int = 0
        """The count of private bookmarks."""
        public: int = 0
        """The count of public bookmarks."""
        untagged: int = 0
        """The count of untagged bookmarks."""
        tagged: int = 0
        """The count of tagged bookmarks."""

    @property
    def counts(self) -> Counts:
        """The various counts for the bookmarks."""
        # TODO: cache this.
        counts = self.Counts(all=len(self._bookmarks))
        for bookmark in self._bookmarks:
            if bookmark.to_read:
                counts.unread += 1
            else:
                counts.read += 1
            if bookmark.shared:
                counts.public += 1
            else:
                counts.private += 1
            if bookmark.tags:
                counts.tagged += 1
            else:
                counts.untagged += 1
        return counts

    @property
    def tags(self) -> list[str]:
        """All known tags used in the bookmarks."""
        tags: set[str] = set()
        for bookmark in self._bookmarks:
            tags |= set(tag for tag in bookmark.tag_list)
        return sorted(list(tags), key=str.casefold)

    def __len__(self) -> int:
        """Returns the number of bookmarks."""
        return len(self._bookmarks)

    def __bool__(self) -> bool:
        """`True` if there are bookmarks, `False` if not."""
        return bool(len(self))

    def __getitem__(self, index: str | int) -> BookmarkData:
        """Get a bookmark.

        Args:
            index: Either the hash of the bookmark, or its integer index.

        Returns:
            The bookmark data.
        """
        if isinstance(index, str):
            return self[self._index[index]]
        return self._bookmarks[index]

    def __iter__(self) -> Iterator[BookmarkData]:
        """An iterator over the bookmarks."""
        return iter(self._bookmarks)


### bookmarks.py ends here



"""Provides a function for checking an URL in the wayback machine."""

##############################################################################
# Python imports.
from json import loads
from typing import Final, NamedTuple

##############################################################################
# HTTPX imports.
import httpx

##############################################################################
# Local imports.
from ..pinboard import API


##############################################################################
class Availability(NamedTuple):
    """The availability data for a URL in the wayback machine."""

    available: bool
    """Is the URL available in the archive?"""

    url: str
    """The URL that was checked."""

    archive_url: str
    """The URL that is available in the archive."""

    timestamp: str
    """The timestamp of the archive entry."""

    status: str
    """The status in the archive."""


##############################################################################
API_URL: Final[str] = "http://archive.org/wayback/available?url="
"""The URL of the API for getting availability."""


##############################################################################
class WaybackError(Exception):
    """Error raised when there is a problem connecting to the wayback machine."""


##############################################################################
async def availability(url: str) -> Availability:
    """Check if a URL is available in the wayback machine.

    Args:
        url: The URL to check.

    Returns:
        An `Availability` object.
    """
    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(
                f"{API_URL}{url}",
                headers={"user-agent": API.AGENT},
            )
        except httpx.RequestError as error:
            raise WaybackError(str(error))
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as error:
            raise WaybackError(str(error))

        result = loads(response.text)
        if "archived_snapshots" in result and "closest" in result["archived_snapshots"]:
            return Availability(
                available=True,
                url=url,
                archive_url=result["archived_snapshots"]["closest"]["url"],
                timestamp=result["archived_snapshots"]["closest"]["timestamp"],
                status=result["archived_snapshots"]["closest"]["status"],
            )
        return Availability(
            available=False,
            url=url,
            archive_url="",
            timestamp="",
            status="",
        )


### api.py ends here
